package net.pricefx.common.util;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.time.DateFormatUtils;
import org.apache.commons.lang3.time.FastDateFormat;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;


public class CalendarUtil {

	private static final Logger log = LoggerFactory.getLogger(CalendarUtil.class);
	
	private static final String DATETIME_PATTERN = DateFormatUtils.ISO_DATETIME_FORMAT.getPattern();
	private static final FastDateFormat DEFAULT_DATETIME_FORMAT = FastDateFormat.getInstance(DATETIME_PATTERN);
	private static final FastDateFormat DEFAULT_DATE_FORMAT = DateFormatUtils.ISO_DATE_FORMAT;
	public static final FastDateFormat FILE_DATE_FORMAT = FastDateFormat.getInstance("yyyyMMdd");
	public static final FastDateFormat FILE_DATETIME_FORMAT = FastDateFormat.getInstance("yyyyMMdd_HHmmss");
	public static final String SIMPLEDATEFORMAT_PATTERN = "EEE MMM dd HH:mm:ss zzz yyyy";

	private FastDateFormat dateFormat;
	private FastDateFormat dateTimeFormat;
	
	
	public enum TimeUnit { 
		DAY("Day"), 
		WEEK("Week"), 
		MONTH("Month"), 
		QUARTER("Quarter"), 
		YEAR("Year");
		
		private String name;
		
		private TimeUnit(final String name){ 
			this.name = name; 
		}

		public String getName(){
			return name;
		}
		
		public Date startDate(final Date date){
			if (date != null){
				final Calendar cal = Calendar.getInstance();
				cal.setTime(date);
				int year = cal.get(Calendar.YEAR);
				int month = cal.get(Calendar.MONTH);
				int week = cal.get(Calendar.WEEK_OF_YEAR);
				int day = cal.get(Calendar.DAY_OF_MONTH);
				
				cal.clear();
				cal.set(Calendar.YEAR, year);		// year is always set
				
				switch(this){
				case DAY: 
					cal.set(Calendar.MONTH, month);
					cal.set(Calendar.DAY_OF_MONTH, day);
					break;
				case WEEK:
					cal.set(Calendar.WEEK_OF_YEAR, week);
					break;
				case MONTH:
					cal.set(Calendar.MONTH, month);
					break;
				case QUARTER:
					month = (month/3)*3;		// month=0..11  -->  quarter start month = [0,3,6,9] 
					cal.set(Calendar.MONTH, month);
					break;
				case YEAR:
					break;
				}
				return cal.getTime();
			}
			return null;
		}

		public Date endDate(final Date date){
			if (date != null){
				final Date endDate = offsetDate(startDate(date), +1);		// 1st day of next period
				return TimeUnit.DAY.offsetDate(endDate,-1);		// the day before that is our end date
			}
			return null;
		}

		public Date offsetDate(final Date date, int offset){
			if (date != null && offset != 0){
				final Calendar cal = Calendar.getInstance();
				cal.setTime(date);
				switch(this){
				case DAY: 
					cal.add(Calendar.DATE, offset);
					break;
				case WEEK:
					cal.add(Calendar.WEEK_OF_YEAR, offset);
					break;
				case MONTH:
					cal.add(Calendar.MONTH, offset);
					break;
				case QUARTER:
					cal.add(Calendar.MONTH, offset*3);
					break;
				case YEAR:
					cal.add(Calendar.YEAR, offset);
					break;
				}
				return cal.getTime();
			}
			return date;
		}
		
		@Override
		public String toString(){
			return name;
		}
		
		public static TimeUnit get(final String unit, TimeUnit defaultValue) {
			try {
				return valueOf(unit==null ? null : unit.toUpperCase());
			} catch(Exception e) {
				log.warn("TimeUnit.get: not a valid unit: {}", unit);
				return defaultValue;
			}
		}
		
		public static TimeUnit get(final String unit) {
			return get(unit, null);
		}
	}
	
	
	public static class TimePeriod {
		@JsonProperty private TimeUnit timeUnit;
		@JsonProperty private Date startDate;
		@JsonProperty private Date endDate;
		@JsonProperty private boolean single;
		private TimePeriod startPeriod;		// getters written so that we don't need to (de-)serialize these
		private TimePeriod endPeriod;
		
		protected TimePeriod() {}  // for json deserializtion
		
		// relying on a proper start date being passed in (hence the protected constructors)
		public TimePeriod(final Calendar cal, final TimeUnit unit){
			this(cal.getTime(), unit);
		}
		
		public TimePeriod(final String timePeriodAsString) throws ParseException{
			final CalendarUtil cal = new CalendarUtil();
			final TimePeriod period = cal.parseTimePeriod(timePeriodAsString);
			this.timeUnit = period.timeUnit;
			this.startDate = period.startDate;
			this.endDate = period.endDate;
			this.single = period.single;
		}
		
		public TimePeriod(final Date startDate, final TimeUnit unit){
			this.timeUnit = unit;
			this.single = true;
			this.startDate = unit.startDate(startDate);
			this.endDate = unit.endDate(startDate);
		}

		public TimePeriod(final Date startDate, final Date endDate, final TimeUnit unit){
			this.timeUnit = unit;
			this.startDate = unit.startDate(startDate);
			this.endDate = unit.endDate(endDate);
		}

		public TimePeriod(final TimePeriod startPeriod, final TimePeriod endPeriod, final TimeUnit unit){
			this(startPeriod.getStartDate(), endPeriod.getEndDate(), unit);
		}

		public Date getStartDate() {
			return startDate;
		}

		public TimeUnit getTimeUnit() {
			return timeUnit;
		}

		public Date getEndDate(){
			return endDate;
		}
		
		@JsonIgnore
		public TimePeriod getStartPeriod() {
			if (startPeriod == null){
				startPeriod = single? this : new TimePeriod(startDate, timeUnit);
			}
			return startPeriod;
		}

		@JsonIgnore
		public TimePeriod getEndPeriod() {
			if (endPeriod == null){
				endPeriod = single? this : new TimePeriod(endDate, timeUnit);
			}
			return endPeriod;
		}
		
		@JsonIgnore
		public TimePeriod add(Integer numberPeriods){
			if (numberPeriods != null){
				if (single){
					return new TimePeriod(timeUnit.offsetDate(startDate, numberPeriods), timeUnit);
				}
				else{
					return new TimePeriod(timeUnit.offsetDate(startDate, numberPeriods), timeUnit.offsetDate(endDate, numberPeriods), timeUnit);
				}
			}
			return null;
		}
		
		@Override
		public String toString(){
			String s = timeUnit.name().substring(0,1);
			if (!single) s+= "S";
			s += "[" + DEFAULT_DATE_FORMAT.format(startDate) + ", " + DEFAULT_DATE_FORMAT.format(endDate) + "]";
			return s;
		}

	}
	
	
	
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Main class
	//
	
	public CalendarUtil(){}
	
	public CalendarUtil(final FastDateFormat dateFormat, final FastDateFormat dateTimeFormat){
		this.dateFormat = dateFormat;
		this.dateTimeFormat = dateTimeFormat;
	}

	public FastDateFormat getDateFormat(){
		return dateFormat == null? getDefaultDateFormat() : dateFormat;
	}

	public FastDateFormat getDateTimeFormat(){
		return dateTimeFormat == null? getDefaultDateTimeFormat() : dateTimeFormat;
	}

	public static FastDateFormat getDefaultDateFormat(){
		return DEFAULT_DATE_FORMAT;
	}

	public static FastDateFormat getDefaultDateTimeFormat(){
		return DEFAULT_DATETIME_FORMAT;
	}

	public String getYear(final Date date){
		if (date != null){
			Calendar cal = Calendar.getInstance();
			cal.setTime(date);
			return String.format("%s", cal.get(Calendar.YEAR));
		}
		return null;
	}

	public String getQuarter(final Date date){
		if (date != null){
			Calendar cal = Calendar.getInstance();
			cal.setTime(date);
			return String.format("%s-Q%d", cal.get(Calendar.YEAR), cal.get(Calendar.MONTH)/3+1);		
		}
		return null;
	}
	
	public String getMonth(final Date date){
		if (date != null){
			Calendar cal = Calendar.getInstance();
			cal.setTime(date);
			return String.format("%s-M%02d", cal.get(Calendar.YEAR), cal.get(Calendar.MONTH)+1);
		}
		return null;
	}
	
	public String getWeek(final Date date){
		if (date != null){
			Calendar cal = Calendar.getInstance();
			cal.setTime(date);
			return String.format("%s-W%02d", cal.getWeekYear(), cal.get(Calendar.WEEK_OF_YEAR));		
		}
		return null;
	}

	public String getDay(final Date date){
		if (date != null){
			return getDateFormat().format(date);
		}
		return null;
	}

	public String today(){
		return getDay(new Date());
	}

	public static String date(){
		return DEFAULT_DATE_FORMAT.format(new Date());
	}
	
	public static String time(){
		return DEFAULT_DATETIME_FORMAT.format(new Date());
	}


	public TimePeriod getTimePeriod(final String periodAsString) {
		if (periodAsString != null){
			try{
				return parseTimePeriod(periodAsString);
			}
			catch(ParseException pe){
				log.debug("getTimePeriod: {}", pe.getMessage());
			}
		}
		return null;
	}

	public TimePeriod newTimePeriod(final String startDateAsString, final String endDateAsString, final String unitAsString) {
		if (startDateAsString != null){
			final Date startDate = parseDate(startDateAsString);
			final Date endDate = endDateAsString == null? null : parseDate(endDateAsString);
			final TimeUnit unit = TimeUnit.get(unitAsString, TimeUnit.DAY);
			return new TimePeriod(startDate, endDate, unit);
		}
		return null;
	}

	public TimePeriod newTimePeriod(final String startDateAsString, final String unitAsString) {
		if (startDateAsString != null){
			final Date startDate = parseDate(startDateAsString);
			final TimeUnit unit = TimeUnit.get(unitAsString, TimeUnit.DAY);
			return new TimePeriod(startDate, unit);
		}
		return null;
	}

	// badly named method, but preserving in case it's used in groovy logic.
	// newTimePeriod(..) is better, but not the exact same implementation.
	public TimePeriod getTimePeriodByDate(String dateAsString, final String unitAsString){	
		final Date date = parseDate(dateAsString);
		final TimeUnit unit = TimeUnit.get(unitAsString);
		return new TimePeriod(date, unit);
	}
	
	public String getTimePeriodName(final TimePeriod period){
		if (period != null){
			return getTimePeriodName(period.getStartDate(), period.getTimeUnit());
		}
		return null;
	}

	public String getTimePeriodName(final Date date, final TimeUnit unit){
		if (date != null && unit != null){
			switch(unit){
			case DAY: 
				return getDay(date);
			case WEEK:
				return getWeek(date);
			case MONTH:
				return getMonth(date);
			case QUARTER:
				return getQuarter(date);
			case YEAR:
				return getYear(date);
			}
		}
		return null;
	}

	public List<String> getTimePeriodNames(final String startPeriodAsString, final String endPeriodAsString, final String unitAsString){
		 List<String> periodNames = new ArrayList<>();
		 List<TimePeriod> timePeriods = getTimePeriods(startPeriodAsString, endPeriodAsString, unitAsString);
		 if (timePeriods != null){
			 for (TimePeriod period : timePeriods){
				 periodNames.add(getTimePeriodName(period));
			 }
		 }
		 return periodNames;
	}


	public List<TimePeriod> getTimePeriods(final String startPeriodAsString, final String endPeriodAsString, final String unitAsString){
		try{
			TimePeriod startPeriod = parseTimePeriod(startPeriodAsString);
			if (startPeriod != null){
				TimePeriod endPeriod = parseTimePeriod(endPeriodAsString);
				if (endPeriod != null){
					TimeUnit unit = TimeUnit.get(unitAsString);
					if (unit != null){
						return getTimePeriods(startPeriod.startDate, endPeriod.endDate, unit);
					}
				}
			}
		}
		catch(ParseException pe){
			log.info("getPeriods: parsing error: {}", pe.getMessage());
		}
		return null;
	}

	public List<TimePeriod> getTimePeriods(final Date startDate, final Date endDate, final TimeUnit unit){
		List<TimePeriod> periods = new ArrayList<>();
		if (startDate != null && endDate != null && unit != null){
			Date date = startDate;
			while (date.compareTo(endDate) <= 0){
				TimePeriod period = new TimePeriod(date, unit);
				if (!periods.contains(period)){
					periods.add(period);
				}
				date = unit.offsetDate(date,+1);
			}
		}
		return periods;
	}
	
	
	public Date getStartDate(final String period){
		if (period != null){
			try{
				return parseTimePeriod(period).getStartDate();
			}
			catch(ParseException pe){
				log.debug("getStartDate: {}", pe.getMessage());
			}
		}
		return null;
	}

	public Date getEndDate(final String period){
		if (period != null){
			try{
				return parseTimePeriod(period).getEndDate();
			}
			catch(ParseException pe){
				log.debug("getEndDate: {}", pe.getMessage());
			}
		}
		return null;
	}
	
	

	// parses the period as a year, quarter, month, week or day
	// them offsets the start date of the resulting time period with the #unit specified by (number, unit)
	// returns the resulting date in the same time unit as the input period
	//
	// ex:  add(2013, 3, MONTH) = 2013
	//      add(2013, -3, MONTH) = 2012
	//      add(2013-M01, 3, MONTH) = 2013-M04

	public String add(final String period, int number, final String unit) {
		if (period != null && unit != null){
			try{
				final TimePeriod timePeriod = parseTimePeriod(period);
				final TimeUnit timeUnit = TimeUnit.valueOf(unit.toUpperCase());
				final Date offsetDate = timeUnit.offsetDate(timePeriod.getStartDate(), number);
				
				switch(timePeriod.getTimeUnit()){
				case YEAR:
					return getYear(offsetDate);
				case QUARTER:
					return getQuarter(offsetDate);
				case MONTH:
					return getMonth(offsetDate);
				case WEEK:
					return getWeek(offsetDate);
				case DAY:
					return getDateFormat().format(offsetDate);
				}
			}
			catch(ParseException pe){
				log.debug("add: {}", pe.getMessage());
			}
		}
		return null;
	}

	public long getDaysDiff(final String period2, final String period1){
		if (period2 != null && period1 != null){
			try{
				final TimePeriod timePeriod2 = parseTimePeriod(period2);
				final TimePeriod timePeriod1 = parseTimePeriod(period1);
				// time diff in msecs / #msecs in day
				// +1 since we want endDate to be inclusive getDaysDiff(2015-M01, 2015-M01) = 31 days (in Jan), not 30 days  
				return (timePeriod2.getEndDate().getTime() - timePeriod1.getStartDate().getTime()) / (1000*60*60*24) +1;  
				
			}
			catch(ParseException pe){
				log.debug("diff: {}", pe.getMessage());
			}
		}
		return 0;
	}

	public TimePeriod parseTimePeriod(final String period) throws ParseException {
		if (period != null) try{
			// try our (custom) year, quarter, month, week formats 
			if (period.length() >= 4){
				try{
					final int year = Integer.parseInt(period.substring(0,4));
					if (year>=1970){
						Calendar cal = Calendar.getInstance();
						if (period.length()==4){
							cal.set(year,0,1, 0,0,0);
							return new TimePeriod(cal, TimeUnit.YEAR);
						}
						else if (period.length()>=6 && period.charAt(4)=='-'){
							final char unit = period.charAt(5);
							// if not a digit then we expect Q,M or W
							if (!Character.isDigit(unit)){
								int n = Integer.parseInt(period.substring(6));
								if (unit=='Q' && n>=1 && n<=4){
									cal.set(year, (n-1)*3, 1, 0,0,0);	// 1st day of the quarter
									return new TimePeriod(cal, TimeUnit.QUARTER);
								}
								else if (unit=='M' && n>=1 && n<=12){
									cal.set(year, n-1, 1, 0,0,0);	// 1st day of the month
									return new TimePeriod(cal, TimeUnit.MONTH);
								}
								else if (unit=='W' && n>=1 && n<=52){
									cal.set(year,0,1, 0,0,0);
									cal.add(Calendar.WEEK_OF_YEAR, n-1);
									return new TimePeriod(cal, TimeUnit.WEEK);
								}
							}
						}
					}
				}
				catch(NumberFormatException ignore){}
					
				// try the standard date and datetime formats 
				Date date = parseDate(period);
				if (date == null) {
					DateTime dateTime = parseDateTime(period);
					if (dateTime != null) date = dateTime.toDate();
				}
				if (date != null) {
					return new TimePeriod(date, TimeUnit.DAY);
				}
			}
		}
		catch(IllegalArgumentException e) { 
			// we throw a parse exception below, if anything else fails
		}
		// nothing worked - invalid period
		throw new ParseException(period,0);
	}
	
	
	public Date parseDate(final String dateAsString) {
		try {
			return getDateFormat().parse(dateAsString);
		} catch (ParseException e) {
			// Continue to below
		}

		try {
			return new SimpleDateFormat(SIMPLEDATEFORMAT_PATTERN).parse(dateAsString);
		} catch (ParseException e) {
			// Continue to below
		}

		// if not a nice and clean date format, try the flexible datetime parser below.
		DateTime dt = parseDateTime(dateAsString);
		if (dt != null) {
			return dt.toDate();
		}

		return null;
	}
	
	
	public DateTime parseDateTime(final String datetimeAsString){
		DateTime datetime = null;
		try{
			final Date date = getDateTimeFormat().parse(datetimeAsString);
			datetime = new DateTime(date);
		}
		catch (ParseException | IllegalArgumentException e){
			// not the standard ISO date format...
			// date seems been passed in some special format. let's use pojava's robust datetime parser
			// datetime seems been passed in some special format. let's use pojava's robust datetime parser
			try{
				org.pojava.datetime.DateTime pojojavaDateTime = org.pojava.datetime.DateTime.parse((String) datetimeAsString);
				datetime = new DateTime(pojojavaDateTime.toDate());
			} 
			catch(Exception e3){
				DateTime jodaDateTime = null;
				try{
					jodaDateTime = DateTime.parse(datetimeAsString);
					if (jodaDateTime == null) throw new IllegalArgumentException();
					datetime = jodaDateTime;
				} 
				catch(Exception e2){	// was IllegalArgumentException but these libs seem to throw all sort of exceptions when not happy with the input
					log.debug("parseDate: invalid datetime '{}': {}", datetimeAsString, e3.getMessage());
				}
			}
		}
		return datetime;
	}

}
