package net.pricefx.server.dto.calculation;

import java.util.*;

import net.pricefx.common.api.CalculationResultType;
import net.pricefx.common.api.InputButtonAction;
import net.pricefx.common.api.InputType;
import net.pricefx.lmformulaengine.LmFormulaEngineContext;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;


/**
 * This class is used to transport a configurator entry.
 *
 * Example:
 * <pre>{@code
 * def opts = ["Europe", "USA"]
 * def configurationEntry = api.createConfiguratorEntry(InputType.OPTION, "Sales_Org_inline_01")
 * def contextParameter = configurationEntry.getFirstInput()
 * contextParameter.setValueOptions(opts)
 * }</pre>
 *
 * @see PublicGroovyAPI#createConfiguratorEntry(InputType, String)
 *
 * @author cat
 *
 */
@JsonIgnoreProperties(ignoreUnknown=true)
public class ConfiguratorEntry implements ComplexCalculationResult {

    private static final long serialVersionUID = -5268805022880613L;

    private List<ContextParameter> inputs;
	private String message;
	private List<InputButtonAction> hiddenActions;
	private transient LmFormulaEngineContext ctx;
	private Map<String,String> additionalConfig;

	/**
	 * Creates an empty configurator entry object. Must be filled with input parameters by calling createParameter on the object.
	 * Example:
	 * <pre>{@code
	 * def cStartDate = api.input("CStartDate")
	 * def configurationEntry = api.createConfiguratorEntry().createParameter(startDate, {cStartDate})
	 * }</pre>
	 */
	public ConfiguratorEntry() {}

	/**
	 * Gets the list of context parameters parameters.
	 * @return the list of context parameters, or null if no context parameters were previously added
	 */
	public List<ContextParameter> getInputs() {
		return inputs;
	}

	/**
	 * Gets the first context parameter.
	 * Example:
	 * <pre>{@code
	 * api.logInfo("fist input value",configuratorEntry.getFirstInput().getValue())
	 * }</pre>
	 * @return the first context parameter, or null if no context parameter was previously added
	 */
	public ContextParameter getFirstInput() {
		return inputs != null ? inputs.get(0) : null;
	}

	/**
	 * Sets the list of context parameters
	 * @see ContextParameter
	 * @param inputs a list of context parameters. Can be null.
	 */
	public void setInputs(List<ContextParameter> inputs) {
		this.inputs = inputs;
	}

	/**
	 * Gets the message
	 * @return the message, or null if it was not set
	 */
	public String getMessage() {
		return message;
	}

	/**
	 * Sets a message that will appear on top of the configuration entry. You may use HTML markup.
	 * Example:
	 * <pre>{@code
	 * entry.setMessage("<BR /><Center><B><Font Size='3'>Competitor Price Information</Font></B></Center></BR/ >")
	 * }</pre>
	 *
	 * @param message the text to display as message
	 */
	public void setMessage(String message) {
		this.message = message;
	}

	public void setHiddenActions(InputButtonAction... hiddenActions) {
		this.hiddenActions = hiddenActions == null ? null : Arrays.asList(hiddenActions);
	}

	public List<InputButtonAction> getHiddenActions() {
		return hiddenActions;
	}

	/**
	 * Creates a parameter and adds it to the list of inputs. The parameter will be created without target date.
	 * If the FormulaEngineContext is set, it will be used to set the value of the context parameter
	 * by retrieving a parameter from the context with the specified name.
	 *
	 * Example:
	 * <pre>{@code
	 * Date targetDate = context == null? new Date() : context.getTargetDate();
	 * ContextParameter param = ce.createParameter(InputType.OPTION, "Customer_inline_01");
	 * }</pre>
	 * @param type Input type of the parameter.
	 * @param name Name of the parameter.
	 * @return the context parameter
	 */
	public ContextParameter createParameter(InputType type, String name) {
		return createParameter(type, name, null);
	}

	/**
	 * Creates a parameter and adds it to the list of inputs.
	 * If the FormulaEngineContext is set, it will be used to set the value of the context parameter
	 * by retrieving a parameter from the context with the specified name.
	 * @see PublicGroovyAPI#createConfiguratorEntry(InputType, String, Date)
	 *
	 * Example:
	 * <pre>{@code
	 * Date targetDate = context == null? new Date() : context.getTargetDate();
	 * ContextParameter param = ce.createParameter(InputType.OPTION, "Customer_inline_01", targetDate);
	 * }</pre>
	 * @param type Input type of the parameter.
	 * @param name Name of the parameter.
	 * @param targetDate Target date for the input parameter (not applicable to all input types).
	 * @return the context parameter
	 */
	public ContextParameter createParameter(InputType type, String name, Date targetDate) {
		ContextParameter input = new ContextParameter(type, name, targetDate, null);
		
		if(ctx!= null) input.setValue(ctx.getParameter(name));

		if(inputs == null) inputs = new ArrayList<>();
		inputs.add(input);

		return input;
	}

	/**
	 * Creates a parameter and adds it to the list of inputs.
	 * If the FormulaEngineContext is set, it will be used to set the value of the context parameter
	 * by retrieving a parameter from the context with the specified name.
	 * If the FormulaEngineContext is not set, or no parameter that matches the name of the input is found
	 * in the context, then defaultValue from the closure will be used.
	 *
	 * Example:
	 * <pre>{@code
	 * ContextParameter annualRebate = configuratorEntry.createParameter(InputType.INTEGERUSERENTRY, "AnnualRebate")
	 * }</pre>
	 * @param input the input element to add
	 * @param defaultValue the value to be used if nothing is selected in the input
	 * @return this ConfiguratorEntry object instance
	 */
	public ConfiguratorEntry createParameter(ContextParameter input, groovy.lang.Closure defaultValue) {
		if (inputs == null) {
			inputs = new ArrayList<>();
		}

		setContextParameterValue(inputs, input, defaultValue);
		return this;
	}

	private void setContextParameterValue(List<ContextParameter> inputs, ContextParameter input, groovy.lang.Closure defaultValue) {
		if (!input.getInputs().isEmpty()) {
			for (ContextParameter nestedInput : input.getInputs()) {
				setContextParameterValue(new ArrayList<>(), nestedInput, defaultValue);
			}
		}

		if (ctx != null) {
			input.setValue(ctx.getParameter(input.getName()));
		}
		if (input.getValue() == null && defaultValue != null) {
			input.setValue(defaultValue.call());
		}

		inputs.add(input);
	}

	/**
	 * Creates a parameter and adds it to the list of inputs.
	 * If the FormulaEngineContext is set, it will be used to set the value of the context parameter
	 * by retrieving a parameter from the context with the specified name.
	 *
	 * Example:
	 * <pre>{@code
	 * def annualRebate = configuratorEntry.createParameter(InputType.INTEGERUSERENTRY)
	 * }</pre>
	 * @param input the input element to add
	 * @return this ConfiguratorEntry object instance
	 */
	public ConfiguratorEntry createParameter(ContextParameter input) {
		return createParameter(input, null);
	}

	/**
	 * Adds a parameter to the list of parameters of this object.
	 * Example:
	 * <pre>{@code
	 * ContextParameter contextParameter = api.inputBuilderFactory()
	 *                                        ... (other methods from the builder)
	 *                                       .buildContextParameter()
	 * ConfiguratorEntry configuratorEntry = api.createConfiguratorEntry().addParameter(contextParameter)
	 * }</pre>
	 * @param input The ContextParameter to add (f.e an input)
	 * @return this ConfiguratorEntry object instance
	 */
	public ConfiguratorEntry addParameter(ContextParameter input) {
		if(inputs == null) inputs = new ArrayList<>();
		inputs.add(input);
		return this;
	}

	/**
	 * Gets the Map of additional configurations
	 * @return the additional config, or null, if nothing was set
	 */
	public Map<String, String> getAdditionalConfig() {
		return additionalConfig;
	}

	/**
	 * Sets all the additional configurations at once
	 * @see ConfiguratorEntry#addAdditionalConfigValue(String, String)
	 * @param additionalConfig a map of additional configuration
	 */
	public void setAdditionalConfig(Map<String, String> additionalConfig) {
		this.additionalConfig = additionalConfig;
	}

	/**
	 * Sets an additional configuration pair
	 * Example:
	 * <pre>{@code
	 * def invoicePriceBoundaries = api.createConfiguratorEntry()
	 * invoicePriceBoundaries.addAdditionalConfigValue("groupLabel", "Business Rules")
	 * invoicePriceBoundaries.addAdditionalConfigValue("showInSectionsPanel", "true")
	 * }</pre>
	 * @param name the name of the configuration key
	 * @param value the value of the configuration key
	 */
	public void addAdditionalConfigValue(String name, String value) {
		if(additionalConfig == null)
			additionalConfig = new HashMap<>();
		additionalConfig.put(name,value);
	}

	/**
	 * The CalculationResultType for this class.
	 * @return CONFIGURATORENTRY
	 */
	@Override
	public String getResultType() {
		return CalculationResultType.CONFIGURATORENTRY.name();
	}

	/**
	 * Sets the FormulaEngineContext. It is use for retrieving values when creating parameters
	 * @param ctx
	 */
	@JsonIgnore
	public void setContext(LmFormulaEngineContext ctx) {
		this.ctx = ctx;		
	}
}
