package net.pricefx.server.dto.calculation;

import net.pricefx.common.api.CalculationResultType;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class ResultHighchart implements ComplexCalculationResult {

	private static final long serialVersionUID = 7030L; // picked Highchart release + 1 digit
	
	private final Map<String, ?> chartDefinition;
	
	private final List<String> modulesList = new ArrayList<>();
	
	private boolean showDataTab = false;
	
	private List<String> headersList = null;
	
	private List<ResultMatrix> resultMatrixList = null;
	
	/**
	 * Not intended for public use. Using this constructor will not fully initialize the instance,
	 * please prefer {@code api.buildHighchart()}.
	 * 
	 * @see PublicGroovyAPI#buildHighchart
	 */
	public ResultHighchart(final Map<String, ?> definition) {
		this.chartDefinition = definition;
	}
	
	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public Map<String, ?> getChartDefinition() {
		return chartDefinition;
	}

	/**
	 * By default, these modules are already available: boost, drilldown, exporting, export-data, heatmap,
	 * no-data-to-display, treemap. If your chart requires additional modules, they can be specified here.
	 * The application ensures the files load only once.
	 * <p>
	 * Note regarding the "boost" module: The <a href="https://api.highcharts.com/highcharts/boost">Highcharts
	 * documentation</a> states that it is enabled by default, but in Pricefx you will need to explicitly set
	 * the option <code>boost.enabled = true</code>
	 * 
	 * @param modules    one or more Highcharts module names, without the path and without the ".js" extension.
	 *                   e.g. "annotations", "funnel", "variwide"
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a> to find out what modules
	 *      are needed for each chart type.
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart addModule(String... modules) {
//		modulesList.addAll(Collections.arrayToList(modules));
		return this;
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<String> getModulesList() {
		return modulesList;
	}
	
	////////////
	// DATA TAB
	////////////
	
	/**
	 * Show a tab set, with the chart in the first tab, and data grids in the other tabs.
	 * By default, a single grid will be shown by extracting the data using the native JavaScript
	 * Highcharts API. The headers of this grid can be customized via {@code setDataHeaders()},
	 * or the whole grid can be provided in a {@link ResultMatrix} instance via
	 * {@code setResultMatrixTabs()}.
	 * 
	 * Note that for convenience, calling {@code setDataHeaders()}, {@code setResultMatrixTabs()},
	 * or {@code addResultMatrixTab()}
	 * will set this to true.
	 * 
	 * @param show    Data tab will be shown if set to true. Default is false.
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setDataHeaders
	 * @see #setResultMatrixTabs
	 * @see #addResultMatrixTab
	 */
	public ResultHighchart setShowDataTab(final boolean show) {
		showDataTab = show;
		return this;
	}
	
	/**
	 * Short for {@code setShowDataTab(true)}
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	public ResultHighchart showDataTab() {
		return setShowDataTab(true);
	}
	
	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public boolean getShowDataTab() {
		return showDataTab;
	}

	/**
	 * To override the default column headers of the data tab table.
	 * 
	 * For convenience, this will also call {@code setShowDataTab(true)}
	 * 
	 * @param headers    One or more header labels, left to right
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setDataHeaders(String... headers) {
//		headersList = Collections.arrayToList(headers);
		return showDataTab();
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<String> getDataHeaders() {
		return headersList;
	}
	
	/**
	 * The automatically filled data tab can be replaced by customized ResultMatrix data.
	 * It's even possible to add multiple tabs by providing multiple ResultMatrix instances.
	 * 
	 * Make sure to set a title via {@code ResultMatrix.setTitle()} as this will be used for the
	 * tab label.
	 * 
	 * Note that:
	 * <ul>
	 * <li>This method will overwrite any previously set or added {@code ResultMatrix}</li>
	 * <li>For convenience, this will also call {@code setShowDataTab(true)}</li>
	 * </ul>
	 * 
	 * @param matrices    One or more ResultMatrix instances to display in the tab set
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setShowDataTab
	 * @see {@link ResultMatrix#setTitle}
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setResultMatrixTabs(ResultMatrix... matrices) {
//		resultMatrixList = Collections.arrayToList(matrices);
		return showDataTab();
	}
	
	/**
	 * The automatically filled data tab can be replaced by customized ResultMatrix data.
	 * It's even possible to add multiple tabs by providing multiple ResultMatrix instances.
	 * 
	 * Make sure to set a title via {@code ResultMatrix.setTitle()} as this will be used for the
	 * tab label.
	 * 
	 * Note that:
	 * <ul>
	 * <li>This method will <b>not</b> overwrite previously set or added {@code ResultMatrix}</li>
	 * <li>For convenience, this will also call {@code setShowDataTab(true)}</li>
	 * </ul>
	 * 
	 * @param matrix    Add one ResultMatrix instance to display in its own tab
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setShowDataTab
	 * @see {@link ResultMatrix#setTitle}
	 */
	public ResultHighchart addResultMatrixTab(ResultMatrix matrix) {
		if (resultMatrixList == null) {
			resultMatrixList = new ArrayList<>();
		}
		resultMatrixList.add(matrix);
		return showDataTab();
	}
	
	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<ResultMatrix> getResultMatrixList() {
		return resultMatrixList;
	}

	@Override
	public String getResultType() {
		return CalculationResultType.HIGHCHART.name();
	}

}
